\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Tests of A331692  continued fraction of sqrt(phi)
\\   1, 3, 1, 2, 11, 3, 5
\\

default(recover,0)
default(strictargs,1)
default(parisizemax,1500*10^6);

read("a331692-contfrac-sqrt-phi.gp");

read("../interv.gp");
read("../interv-wip.gp");
read("../test-helpers.gp");
read("OEIS-data.gp");
read("OEIS-data-wip.gp");
option_verbose=0;
option_progress = 1;


\\ cf
\\ A356664  sqrt phi convs
\\ A356591  sqrt phi convs


\\-----------------------------------------------------------------------------
\\ sqrt_phi_interval()

{
  \\ form using two sqrts instead of quad
  my(sqrt_phi_interval_by_sqrts(bits) =
     my(s=1/2^bits,
        x=s*sqrtint( 1<<(2*bits-1) + sqrtint(5<<(4*bits-2)) ));
     [x,x+s]);
  check_equal(vector(500,bits,bits--;
                  sqrt_phi_interval(bits)),
              vector(500,bits,bits--;
                  sqrt_phi_interval_by_sqrts(bits)));
}

sqrt_phi_pol = 'x^4 - 'x^2 - 1;
sqrt_phi_pol_f(x) = subst(sqrt_phi_pol,'x,x);
{
  if(0,
     print("sqrt_phi_pol roots: ");
     foreach(polroots(sqrt_phi_pol),r,
        print("  "r)));
}
{
  \\ 0 crossing by interval
  for(bits=0,1000,
     my(t=sqrt_phi_interval(bits));
     sqrt_phi_pol_f(t[1]) < 0 || error();
     sqrt_phi_pol_f(t[2]) > 0 || error();
     );
}

{
  \\ versus interv_solve_bisection()
  forstep(bits=21,500,15,
     my(by_bisecion = interv_solve_bisection
                      ([1,2], x->subst(sqrt_phi_pol,'x,x),
                       bits),
        by_bits=sqrt_phi_interval(bits));
     check_equal(by_bits,by_bisecion);
  );
}


\\-----------------------------------------------------------------------------
\\ A331692_vector_bits()

{
  my(want=OEIS_data("A331692"),
     got=A331692_vector_bits(1000),
     len=min(#got,#want));
  check_equal(#got >= #want, 1);
  print("A331692 vs DATA got len ",#got," want len ",#want," len ",len);
  check_equal(got [1..len],
              want[1..len]);

  \\ value from contfrac terms, has 0 crossing
  my(t=interv_from_contfrac(got),
     t_subst = interv_subst(sqrt_phi_pol,x,t));
  check_equal(interv_contains_0(t_subst), 1);
}
{
  \\ vs interv_contfrac()
  my(bits=10000,
     by_bits=A331692_vector_bits(bits),
     t=sqrt_phi_interval(bits),
     by_interv=interv_contfrac(t));
  check_equal(by_bits,by_interv,
              "A331692_vector_bits() vs interv_contfrac()");

  \\ 0 crossing
  my(t=interv_from_contfrac(by_bits),
     t_subst = interv_subst(sqrt_phi_pol,x,t));
  check_equal(interv_contains_0(t_subst), 1);

  \\ vs float contfrac()
  localprec(bits+1000);
  my(s=sqrt((1 + sqrt(5))/2),
     by_float=contfrac(s));
  check_equal(by_bits,by_float[1..#by_bits],
              "A331692_vector_bits() vs float contfrac()");
}
{
  \\ make b-file
  if(0,
     my(v=A331692_vector_bits(35000));
     my(at); print("contrfrac len "#v" max ",vecmax(v,&at)," at ",at);
     system("mkdir -p /tmp/new");
     system("rm -f /tmp/new/b331692.txt");
     for(i=1,10001,
        my(n=i-1,term=v[i]);
        write("/tmp/new/b331692.txt",Str(n" "term)));
     system("ls -l /tmp/new/b331692.txt");
     quit);
}
{
  \\ check b-file
  \\ about 35000 bits for 10000 terms
  if(0,
     progress("A331692 b331692.txt");
     my(want=OEIS_data("A331692",'bfile),
        got=A331692_vector_bits(35000),
        len=min(#got,#want));
     check_equal(#got >= #want, 1);
     print("A331692 vs bfile got len ",#got," want len ",#want," len ",len);
     check_equal(got [1..len],
                 want[1..len]);

     \\ value from contfrac terms, has 0 crossing
     my(t=interv_from_contfrac(got),
        t_subst = interv_subst(sqrt_phi_pol,x,t));
     check_equal(interv_contains_0(t_subst), 1);
    );
}


\\ {
\\   my(bitlen=300);
\\   my(t = interv_solve_bisection
\\         ([1,2], x->subst(sqrt_phi_pol,'x,x),
\\          bitlen));
\\   print("den bitlen ",logint(denominator(t[1]),2));
\\   my(got=interv_contfrac(t));
\\   \\ check_equal(#got,309);
\\
\\ }

\\ {
\\   my(want=OEIS_data("A331692"),
\\      got=A331692_vector_bits(1000));
\\
\\
\\   OEIS_check_vecfunc("A331692_vector");
\\   if(0,
\\      OEIS_check_vecfunc("A331692_vector",'bfile));
\\ }


\\-----------------------------------------------------------------------------
\\ A225204 / A225205 Convergents

{
  my(want_nums=OEIS_data("A225204"),
     want_dens=OEIS_data("A225205"));

  my(bits=30000,
     len=max(#want_nums,#want_dens)+100,
     t=sqrt_phi_interval(bits),
     v=interv_contfrac(t,len),
     m=contfracpnqn(v,len));
  check_equal(matsize(m), [2,len]);
  check_equal(m[1,1..#want_nums],
              want_nums,
              "convergent nums");
  check_equal(m[2,1..#want_dens],
              want_dens,
              "convergent dens");

  my(A331692(n) = v[n+1]);
  my(A225204(n) = want_nums[n+1]);
  my(A225205(n) = want_dens[n+1]);
  OEIS_check_func("A331692",A331692);
  OEIS_check_func("A225204",A225204);
  OEIS_check_func("A225205",A225205);

  my(a=A225204);
  \\ formula by Jianing Song in A225204
  check_equal(vector(20,n,n++; a(n)),
              vector(20,n,n++;
                     A331692(n)*a(n-1) + a(n-2)));
}

\\-----------------------------------------------------------------------------
\\ Other

{
  if(0,
     my(bits=1000,
        t=sqrt_phi_interval(bits));
     print("Engel ",interv_Engel_terms(t,15));
     \\ not in OEIS: 1, 4, 12, 18, 41, 42, 94, 162, 252, 275, 323, 1004, 1542, 1923, 4541


     );
}



\\-----------------------------------------------------------------------------
\\ range by sqrts

{
  my(t=sqrt_phi_interval(32));
  check_equal(apply(contfrac,t),
     [[1,3,1,2,11,3,5,1,10,1,1,2,3,4,10,3,4,2,3],
      [1,3,1,2,11,3,5,1,10,1,3,2,2,3,1,3,1,13,2,3,3]]);
}

\\-----------------------------------------------------------------------------

{
  my(bitlen=35000);
  \\ my(t=t_polroot_to_bitlen(sqrt_phi_pol,bitlen));
  my(t=sqrt_phi_interval(bitlen));
  print("den bitlen ",logint(denominator(t[1]),2));

  check_equal(  lex(sqrt_phi_pol_f(t[1]),0)
              * lex(sqrt_phi_pol_f(t[2]),0),
              -1,
             "crosses the root");

  my(got=interv_contfrac(t));
  if(0,
     my(want=OEIS_data("A331692",'bfile),
        len=min(#got,#want));
     print("bfile got len ",#got," want len ",#want," len ",len);
     check_equal(got [1..len],
                 want[1..len]));
}

\\-----------------------------------------------------------------------------
\\ A331692_vector_by_quads()
\\
\\ Much slower than interval.
\\
\\ Floor of x,y wrong, x*d is not an integer.
\\ The frac of y discarded might push to next int.

\\ \\-----------------------------------------------------------------------------
\\ \\ A331692_vector_by_quads()
\\ \\
\\ \\ The following is a much slower approach keeping an exact value for the
\\ \\ working remainder.
\\
\\ my(phi=quadgen(5)); \
\\ A331692_vector_by_quads(len) =
\\ {
\\   my(x=0,   \\ working remainder x + y*sqrt(phi),
\\      y=1);  \\ with x and y type t_QUAD
\\   vector(len,n,
\\          my(d = denominator(x),
\\             yyphi = y^2*phi,
\\             term = (x*d + sqrtint(floor(yyphi*d^2)) )\d);
\\ \\ print("x*d = ",x*d);
\\ print("x*d = ",ceil(x*d) - x*d +0.0);
\\          x -= term;
\\          my(nn = 1/(x^2 - yyphi));
\\          x *= nn;
\\          y *= -nn;
\\          term);
\\ }
\\
\\ {
\\   my(v=A331692_vector_bits(1000));
\\   A331692_vector_by_quads(#v) == v  || error();
\\ }
\\
\\ \\ In A331692_vector_by_quads(), the working remainder is
\\ \\
\\ \\     rem = x + y*sqrt(phi)
\\ \\
\\ \\ where x and y are type t_QUAD (in phi), and so representing an element of
\\ \\ field Q[sqrt(phi)].
\\ \\
\\ \\ Each continued fraction term is
\\ \\
\\ \\     term = floor(rem)        sequence term
\\ \\
\\ \\ which is calculated,
\\ \\
\\ \\                 / x*d  +  sqrt(y^2*d^2*phi) \
\\ \\     term = floor| ------------------------- |
\\ \\                 \            d              /
\\ \\
\\ \\     where d = denominator(x)
\\ \\
\\ \\ x*d is an integer so that the sqrt can round down with sqrtint() without
\\ \\ losing anything in the resulting term.
\\ \\
\\ \\ After subtracting "term", the reciprocal is a usual quadratic conjugate
\\ \\
\\ \\           1            x - y*sqrt(phi)
\\ \\     ---------------  = ---------------
\\ \\     x + y*sqrt(phi)          nn
\\ \\
\\ \\     where
\\ \\     nn = (x + y*sqrt(phi)) * (x - y*sqrt(phi))
\\ \\        = x^2 - y^2*phi
\\ \\
\\ \\ These x and y have rational and irrational parts which are fractions and
\\ \\ which grow steadily, and apparently run much slower than the intervals
\\ \\ method above.
\\ \\
\\ \\ Some nfinit() or t_POLMOD might suit Q[sqrt(phi)] values.  That'd be
\\ \\ convenient for the arithmetic, possibly with a common denominator taken
\\ \\ out.  But care still needed at the floor operation.


\\ {
\\   forstep(bits=27,500,35,
\\      my(by_interval=A331692_vector_bits(bits),
\\         by_quads=A331692_vector_by_quads(#by_interval));
\\      check_equal(by_interval,by_quads,
\\                  Str("bits "bits" is len "#by_interval)));
\\ }

\\     rem' = 1/(rem - term)    new remainder
\\
\\ The floor is by taking out denominator(x) from x and y so
\\
\\     d = denominator(x)
\\
\\                 / x*d  +  y*d*sqrt(phi) \
\\     term = floor| --------------------- |
\\                 \          d            /
\\
\\ x*d is an integer so it suffices to discard any fractional part of
\\ sqrt(y^2*d^2*phi), and for that in turn it suffices to discard any
\\ fractional part of its argument y^2*d^2*phi.
\\
\\ After subtracting term from x, the reciprocal is a usual quadratic
\\ conjugate, in this case happening to sqrt of phi rather than an integer.
\\
\\                   1
\\     1/rem = ---------------
\\             x + y*sqrt(phi)
\\
\\             x - y*sqrt(phi)
\\           = ---------------
\\                   nn
\\
\\     where
\\     nn = (x + y*sqrt(phi)) * (x - y*sqrt(phi))
\\        = x^2 - y^2*phi
\\
\\ The effect of x and y is to calculate in a field Q[sqrt(phi)], with a
\\ basis 1, phi, sqrt(phi), phi*sqrt(phi).  An nfinit() or corresponding
\\ t_POLMOD can make such calculations, but not the floor as such.

\\ x + y*sqrt(1/2 + 1/2*sqrts)
\\ xd + sqrt(yd^2*(1/2 + 1/2*sqrts))
\\ ( 4*xd + sqrt(yd^2*(1 + sqrts)) ) /4d
\\
\\ floor_xysqrt(x,y,s) =
\\ {
\\   \\ FIXME: This depends on x*d being an integer ...
\\   my(d=4*denominator([x,y]));
\\   x*=d; y*=d;
\\   (x + sqrtint(floor(y^2*s))) \ d;
\\ }
\\ {
\\   check_equal(floor_xysqrt(0,1,phi), 1);
\\
\\   my(x=151/59 + 77/59*phi,
\\      y=271/118 + 121/118*phi);
\\   print(x+y*sqrt(phi));
\\   \\ check_equal(floor_xysqrt(x,y,phi), 1);
\\ }



\\-----------------------------------------------------------------------------
\\ Quads Growth of X,Y

{
  if(0,
     my(phi=quadgen(5));
     my(x=0,   \\ working remainder x + y*sqrt(phi)
        y=1);  \\ with x and y type t_QUAD of phi
     for(n=1,10000,
        my(d = denominator(x),
           yyphi = y^2*phi,
           term = (x*d + sqrtint(floor(yyphi*d^2)) )\d);
        x -= term;
        my(nn = x^2 - yyphi);
        x /= nn;
        y /= -nn;
        print("n="n" "term,
              "  ",if(x,logint(floor(numerator(x)),2)),
              " ",logint(denominator(x),2),
              " ",if(y,logint(floor(numerator(y)),2)),
              " ",logint(denominator(y),2));
     );
  quit);
}


\\-----------------------------------------------------------------------------
\\ Unused

\\ ff_floor(v) = v[1] >> v[2];
\\ ff_celi(v) = v[1] >> v[2];
\\ ff_recipfloor(v) = [1 << 2*v[2] \ v[1], v[2]];
\\
\\ finterv_floor(v) =
\\ {
\\   my(ret = v[1] >> v[3]);
\\   if(v[2] >> v[3] == ret,
\\      ret, 'unknown);
\\ }
\\ finterv_recip(v) =
\\ {
\\   my(t = 1<<(2*v[3]),
\\      hi = t \ v[2],
\\      lo = t \ v[1]);
\\   [lo,hi,v[3]];
\\ }
\\ finterv_common(v) =
\\ {
\\   my(d=v[2]-v[1],
\\      k = if(d,logint(d,2)));
\\   [v[1] >> k,
\\    rshift_ceil(v[2],k),
\\    v[3]-k];
\\ }



\\-----------------------------------------------------------------------------
check_equal(i,'i);
check_equal(n,'n);
check_equal(x,'x);
check_equal(y,'y);
check_equal(term,'term);
print("end");
