\\ Copyright 2020, 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, August 2022
\\
\\ This is some PARI/GP code to calculate terms of A331692, the continued
\\ fraction expansion of sqrt(phi), where phi = (sqrt(5)+1)/2 is the golden
\\ ratio.
\\
\\ The strategy is a rational interval surrounding sqrt(phi) and a generic
\\ calculation of continued fraction terms from that.
\\
\\ Requires PARI/GP 2.13 or higher for floor() of t_QUAD.

default(recover,0)
default(strictargs,1)

\\ sqrt_phi_interval(bits) returns a 2 element vector [lo,hi] which are
\\ rationals 1/2^bits apart and bounding lo < sqrt(phi) < hi.
\\
my(phi=quadgen(5)); \
sqrt_phi_interval(bits) =
{
  my(s=1/2^bits,
     lo=s*sqrtint(floor(phi<<(2*bits))));
  [lo, lo+s];
}

\\ contfrac_of_interval(t) takes a vector t = [lo,hi] of rationals lo,hi
\\ which are an interval bounding some constant.
\\ Return a vector of continued fraction terms which are known with
\\ certainty using this interval.
\\
contfrac_of_interval(t) =
{
  my(ret=List([]), f);
  while(f=floor(t); f[1]==f[2],
        listput(ret,f[1]);
        t-=f;
        t=[1/t[2],1/t[1]]);
  Vec(ret);
}

\\ A331692_vector_bits(bits) returns a vector of terms of A331692, up to
\\ those known with certainty within a "bits" many bits calculation.
\\
A331692_vector_bits(bits) = \
  contfrac_of_interval(sqrt_phi_interval(bits));

{
  my(v=A331692_vector_bits(70));
  print1("A331692 = ");
  for(n=1,#v, print1(v[n],","));
  print("...");
  print();
}
{
  \\ print how sqrt_phi_interval() bounds sqrt(phi)
  my(bits=8,
     t=sqrt_phi_interval(bits),
     phi=quadgen(5),
     p);
  print("sqrt_phi_interval("bits") = ",t);
  my(pr(name,x) = x=floor(x);
     printf("  %-9s = %d.%06d...\n",
            name, x\10^6, x%10^6));
  pr("lo",        t[1]*10^6);
  pr("sqrt(phi)", sqrtint(floor(phi*10^12)));
  pr("lo",        t[2]*10^6);
}

\\ sqrt_phi_interval() starts from 2*bits of phi with floor() of a t_QUAD,
\\ and sqrtint() of that is "bits" of sqrt(phi).  This is lo = sqrt(phi)
\\ rounded down to 1/2^bits and the hi bound is then lo + 1/2^bits.
\\
\\ The 2*bits of phi can be formed by a second inner sqrtint() if desired,
\\ but t_QUAD is convenient.
\\
\\ contfrac_of_interval() returns terms in the style of contfrac(), so
\\ ret[1] is the integer term and the t interval has value
\\
\\                         1
\\     t = ret[1] + ------------------------
\\                                1
\\                  ret[2]  + --------------
\\                                      1
\\                            ret[3] + ----
\\                                     etc
\\
\\ While each term found from the lo and hi bounds agree, those terms are
\\ certain.  When lo and hi disagree, that term and all beyond are not
\\ certain.
\\
\\ The rationals within contfrac_of_interval() could be some floating point
\\ binary fractions instead, with suitable rounding direction at each
\\ operation.  They could be decreasing precision too, since there's not
\\ much point keeping many bits below where lo and hi differ.  But leave all
\\ that to a floating point library (with rounding control).
\\
\\ A disadvantage of an interval calculation is that it's not known in
\\ advance how many good contfrac terms will result.  For sqrt(phi), the
\\ average seems around 3.5 bits per term for smallish numbers of terms.


\\-----------------------------------------------------------------------------
print("end");
