\\ Copyright 2020, 2021, 2022, 2023, 2024 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Work in progress.  Will change wildly.  Mostly working but not
\\ particularly good in places.
\\
\\     t = interv_intersect(x,y)     or 'none if disjoint
\\     t = interv_abs(t)
\\     s = interv_sign(t)
\\
\\     t = interv_sqr(t)           square t^2
\\     t = interv_sqrtint(t)
\\     t = interv_sqrtnint(t,n)
\\     cmp = interv_lex(x,y)       or error if cannot cmpare
\\     bool = interv_lt(x,y)       or error if cannot cmpare
\\
\\     f = interv_floor_to_constant(t)   or error if not constant
\\     f = interv_ceil_to_constant(t)
\\     f = interv_round_to_constant(t)
\\     f = interv_floor(t)
\\     c = interv_ceil(t)
\\     t = interv_frac(t)
\\
\\     t = interv_subst(poly,var,t)
\\     t = interv_sumalt(func,num_terms,n_start=0)
\\
\\     vec = interv_Egyptian_terms_alternating(t,maxlen='none)
\\
\\     interv_validate(t)
\\     t = interv_log(t,num_terms)
\\     t = interv_log_1plusx(t,num_terms)
\\     t = interv_exp(t,num_terms)
\\     n = interv_exp_estimate_num_terms(x,bitpos)
\\
\\     t = interv_atan_Taylor(t,terms)
\\
\\ End Summary


\\-----------------------------------------------------------------------------
\\ Sums of Series

\\ Is num_terms sensible, or better a requested width?
\\ When summing to target decimal digits or bits would want
\\ to continue until last digit is certain.

interv_sumalt(func,num_terms,n_start=0) =
{
  num_terms>=2 || error("interv_sumalt() is for num_terms>=2");
  my(t=[0,0],prev,total=0);
  for(n=n_start,n_start+num_terms-1,
     prev=total;
     total+=func(n));
  vecsort([prev,total]);
}
{
addhelp(interv_sumalt,
"t = interv_sumalt(func,num_terms,n_start=0)
Return an interval for sum of func(n) for \"num_terms\" of n starting at
n = n_start.

Each call is term=func(n) to get a term of the sum.  Terms are expected to
be alternating signs and decreasing magnitude so the sums to num_terms and
num_terms-1 are an interval surrounding what would be an infinite sum.");
}


\\-----------------------------------------------------------------------------

interv_validate(t) =
{
  t[1]<=t[2] || error("interval out of order");
}


\\-----------------------------------------------------------------------------
\\ Polynomials

interv_subst(poly,var,t) =
{
  \\ print("p "valuation(p,var)" "poldegree(p,var)" "polcoeff(p,1,var));
  sum(i=valuation(poly,var),poldegree(poly,var),
      vecsort(polcoeff(poly,i,var) * interv_pow_constant(t,i)));
}
{
addhelp(interv_subst,
"t = interv_subst(poly,var,t)
poly is a polynomial in variable \"var\".  t is an interval.
Return the interval which is poly evaluated in the range of t.");
}


\\-----------------------------------------------------------------------------
\\ log and exp

interv_log(t,num_terms) =
{
  \\ log(x) = -log(1/x) = -log(1-x/x+1/x) = -log(1-(x-1)/x)
  \\ log(7)
  \\ -log(1/7)
  \\ -log(1 - 6/7)
  t=vecsort(t);
  if(t[1]>=1,
     interv_neg
     (interv_log_1plusx
      (interv_neg
       (interv_div(interv_add_constant(t,-1), t)),
       num_terms))
     ,
     interv_log_1plusx(interv_add_constant(t,-1), num_terms));
}
{
addhelp(interv_log,
"t = interv_log(t,num_terms)
t is an interval.  Return an interval for log(t).

The current implementation uses Mercator's series with \"num_terms\" many
terms (calculated on the reciprocal if t > 2).");
}

\\ log(1+x)
\\ omit sum i>=n  (-1)^n * x^n/n
\\
interv_log_1plusx(t,num_terms) =
{
  \\ for x<0, i=1 lower bound want even num terms
  \\          i=2 upper bound want odd num terms
  \\
  t=vecsort(t);
  num_terms = ceil(num_terms);
  (-1 < t[1] && t[2] < 1) || error("interv_log_1plusx() is for -1 < t < 1");
  vector(#t,i,
         my(x=-t[i],
            L=if(x<0,
                 num_terms + bitand(1,bitxor(num_terms,i)),
                 num_terms),
            p=-1,
            e=sum(n=1,L-1,
                  \\ print("n="n" term ",(p*x)/n);
                  (p*=x)/n));
         \\ print("i="i" L="L" x="x"  e="e"  ending p="p);
         if(x<0, if(i==1,p<0||error()); if(i==2,p>0||error()));

         if(i==1&&x>0,
            \\ lower bound on negative omitted part
            \\ print("  omitted ",1/(L*(1-x))," * ",p*x);  p<0 || error();
            e + p*x/(L*(1-x)),
            e));

  \\ interv_validate(x);
  \\ \\ x[1]>-1 || error("interv_log_1plusx() must have x>=1");
  \\ my(ret=[0,0]);
  \\ for(i=1,num_terms,
  \\    interv_validate(ret);
  \\    my(t=interv_pow_constant(x,i)/i);
  \\    if(i%2==0, t=interv_neg(t));
  \\    interv_validate(t);
  \\    ret += t);
  \\ ret;
}
{
addhelp(interv_log_1plusx,
"t = interv_log_1plusx(t,num_terms)
Return an interval for log(1+t) of interval t.
This is Mercator's series of num_terms, so down to term 1/(n-1)*x^(n-1).
t must be in the range -1 <= t <= 1 which is the series radius of
convergence, and note that convergence is slow for abs(t) near 1.

For positive values in t, the series alternates and upper and lower bounds
are by stopping at an odd or even term (positive or negative).  This uses 1
term additional to num_terms if necessary.

For negative values in t, a simple negative lower bound is made by a
geometric series for the omitted part.");
}

\\ exp('t)
\\ omit n>=L       sum x^n/n!
\\      <=  x^L/L! sum (x/L)^j
\\ x/L < 1
\\ L > x
\\ suminf(j=0, (2/7)^j)
\\ 1/(1 - 2/7)  *1.0
\\ 1/(1 - x/L)
\\ GP-Test  1/(1 - 'x/'L) == 'L/('L - 'x)
\\
interv_exp(t,num_terms) =
{
  \\ for x<0, i=1 lower bound want even num terms
  \\          i=2 upper bound want odd num terms
  \\
  t=vecsort(t);
  num_terms = ceil(num_terms);
  vector(#t,i,
         my(x=t[i],
            L=if(x<0,
                 num_terms + bitnegimply(1,bitxor(num_terms,i)),
                 max(num_terms, floor(x)+1)),
            p=1,
            e=(L>0) + sum(n=1,L-1, p=p*x/n));
         \\ print("i="i" L="L"  e="e"  ending p="p);
         if(x<0, if(i==1,p<0||error()); if(i==2,p>0||error()));

         if(i==2&&x>0,
            \\ upper bound on omitted part
            \\ print("  omitted ",L/(L - x)," * ",p*x);
            e + p*x/(L - x),
            e));
}
{
addhelp(interv_exp,
"t = interv_exp(t,num_terms)
t is an interval.  Return an interval for exp(t).
Currently this uses the Taylor series expansion, of \"num_terms\" many terms.

For negatives in t, the series alternates and an upper or lower bound
endpoint is found by stopping at an even or odd term as appropriate.  This
may be 1 term additional to num_terms if necessary.

For positives in t, the series is increasing and a simple upper bound is
made by a geometric series for the omitted part.  For hi the maximum in t,
this requires continuing until term n with hi/n < 1, so num_terms is
increased if necessary to at least hi many.");
}

\\ term x^n/n!
interv_exp_estimate_num_terms(x,bitpos) =
{
  x=vecmax(x);
  floor(log(x));
}

\\-----------------------------------------------------------------------------

interv_ceil_to_constant(t) =
{
  t=ceil(t);
  if(t[1] != t[2],
     error("interv_ceil_to_constant(): interval wider than one integer"));
  t[1];
}

interv_floor_to_constant(t) =
{
  t=floor(t);
  if(t[1] != t[2],
     error("interv_floor_to_constant(): interval wider than one integer"));
  t[1];
}
{
addhelp(interv_floor_to_constant,
"f = interv_floor_to_constant(t)
t is an interval.  If its low and high have the same floor, so entirely
between integers i <= t < i+1, then return the integer part i.
If t spans two integer parts then throw an error.");
}

interv_round_to_constant(t) = \
  interv_floor_to_constant(interv_add_constant(t,1/2));
{
addhelp(interv_round_to_constant,
"f = interv_round_to_constant(t)
t is an interval.  If its low and high have the same nearest integer,
so entirely integers i-1/2 <= t < i+1/2, then return the integer part i.
If t spans two rounded integer parts then throw an error.");
}
\\ FIXME: This any good?
interv_floor(t) =
{
  my(ret=floor(vecmin(t)));
  if(vecmax(t)<ret+1, ret, 'none);
}
{
addhelp(interv_floor,
"f = interv_floor(t)
Return floor(t) of an interval t, or return 'none if the accuracy of t is
too wide to know the floor (t spans more than one integer).");
}

interv_ceil(t) =
{
  my(ret=ceil(vecmax(t)));
  if(vecmin(t)>ret-1, ret, 'none);
}
{
addhelp(interv_ceil,
"c = interv_ceil(t)
Return ceil(t) of an interval t, or return 'none if the accuracy of t is
too wide to know the floor (t spans more than one integer).");
}


\\-----------------------------------------------------------------------------

interv_Egyptian_terms_alternating(t,maxlen='none) =
{
  my(ret=List([]));
  t=vecsort(t);
  while(t[1]>0 && (maxlen=='none || #ret<maxlen),
        \\ fractions parts
        my(d=floor(interv_recip(t)));
        d[1]==d[2] || break;
        listput(ret,d[1]);
        t = -interv_sub_constant(t,1/d[1]));
  Vec(ret);
}

\\-----------------------------------------------------------------------------

interv_sqr(t) =
{
  if(interv_contains_0(t),
     [0, sqr(vecmax(abs(t)))],
     vecsort(apply(sqr,t)));
}
{
addhelp(interv_sqr,
"t = interv_sqr(t)
t is an interval.  Return it squared t^2.");
}

interv_sqrtint(t) =
{
  my(r);
  t=interv_widen_integer(t);
  [sqrtint(t[1]),
   sqrtint(t[2],&r) + (r!=0)];
}
{
addhelp(interv_sqrtint,
"t = interv_sqrtint(t)
Return an interval which is the integer sqrts of the given interval t.
This means sqrtint() for the low of t, and and equivalent rounded up to an
integer for the high of t (so +1 when not a perfect square) .

sqrtint() as of PARI 2.13 expects integer operands so do an
interv_widen_integer() first if necessary.  Future PARI might take fractions
in sqrtint() already.");
}

interv_sqrtnint(t,n) =
{
  my(hi=vecmax(t));
  [sqrtnint(vecmin(t),n),
   if(hi>0,
      hi = sqrtnint(ceil(hi)-1,n) + 1,
      sqrtnint(hi,n))];
}
{
addhelp(interv_sqrtnint,
"t = interv_sqrtnint(t,n)
Return an interval which is the integer n'th root of given interval t.
This means sqrtnint(x,n) for the low of t and an equivalent rounded up for
the high of t.

sqrtnint() as of PARI 2.13 expects integer operands so do an
interv_widen_integer() first if necessary.  Future PARI might take fractions
in sqrtnint() making that unnecessary.");
}


\\-----------------------------------------------------------------------------

interv_abs(t) =
{
  t = vecsort(t);
  if(t[1] > 0, t,
     t[2] < 0, - Vecrev(t),
     [0, vecmax(abs(t))]);
}
{
addhelp(interv_abs,
"t = interv_abs(t)
Return absolute value of interval t.");
}

interv_sign(t) =
{
  if(t[1]==0 && t[2]==0, 0,
     t[1]>0 && t[2]>0, 1,
     t[1]<0 && t[2]<0, -1,
     error("interval spans 0"));
}
{
addhelp(interv_sign,
"s = interv_sign(t)
Return sign -1, 0, 1 of the interval t.
-1 means entirely <0, 1 means entirely >0, 0 means exactly 0,
and otherwise throw an error.");
}     

interv_lex(x,y) =
{
  x=vecsort(x);
  y=vecsort(y);
  if(x[2] < y[1], -1,
     x[1]==y[1] && x[2]==y[2], 0,
     x[1] > y[2], 1,
     error("interv_lex() cannot compare overlapping intervals"));
}
interv_lt(x,y) =
{
  my(c=interv_lex(x,y));
  if(c=='unknown, c, c<0);
}

interv_intersect(x,y) =
{
  \\ With swap as necessary so xlo <= ylo,
  \\
  \\  xlo--------------xhi
  \\            ylo-----------yhi
  \\             |-----|
  \\
  \\  xlo---------------------xhi
  \\            ylo----yhi
  \\             |------|
  \\
  \\  xlo--xhi
  \\            ylo----yhi

  x=vecsort(x);
  y=vecsort(y);
  if(x[1] > y[1], [x,y]=[y,x]);
  if(y[1] <= x[2], [y[1], min(x[2],y[2])],
     'none);
}
{
addhelp(interv_intersect,
"t = interv_intersect(x,y)
x and y are intervals.  Return the interval where they overlap,
or return symbol 'none if they don't overlap.");
}

interv_frac(t) =
{
  if(floor(t[1]) != floor(t[2]),
     error("interv_frac(): not enough precision"));
  vecsort(frac(t));
}
{
addhelp(interv_frac,
"t = interv_frac(t)
t is an interval.  If it falls entirely between integers i <= t < i+1,
then return t-i, being the fractional part of t.
If t spans two integers then throw an error.");
}


\\-----------------------------------------------------------------------------

\\ Taylor series, alternating signs
\\                     1       1       1
\\     atan(x) = x - ----- + ----- - ----- + ...
\\                   3*x^3   5*x^5   7*x^7
interv_atan_Taylor(t,terms) =
{
  my(ret=[0,0],prev=ret);
  for(i=1,terms,
     prev=ret;
     ret=interv_add(ret,
       interv_pow_constant(t,2*i-1)*(-1)^(i-1)/(2*i-1)));
  interv_union(prev,ret);
}

\\-----------------------------------------------------------------------------
