\\ Copyright 2020, 2021, 2022, 2023, 2024 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0)
default(strictargs,1)
read("interv.gp");
read("interv-wip.gp");
read("test-helpers.gp");

option_verbose = 0;


\\-----------------------------------------------------------------------------
\\ interv_abs()

{
  check_equal(interv_abs([20,10]), [10,20]);
  check_equal(interv_abs([-10,-20]), [10,20]);
  check_equal(interv_abs([7,-5]), [0,7]);
}


\\-----------------------------------------------------------------------------
\\ interv_sqr()

{
  check_equal(interv_sqr([-3,2]), [0,9]);
  check_equal(interv_sqr([-2,3]), [0,9]);
  check_equal(interv_sqr([0,3]), [0,9]);
  check_equal(interv_sqr([3,4]), [9,16]);
  check_equal(interv_sqr([4,3]), [9,16]);
}

\\-----------------------------------------------------------------------------
\\ interv_atan_Taylor()

{
  check_equal(interv_atan_Taylor([1/2,1/2],1),
             [0, 1/2]);

  check_equal(atan(x+O(x^5)) == x - 1/3*x^3, 1, "atan with O()");
  check_equal(atan(Ser(x,,6)) == x - 1/3*x^3 + 1/5*x^5, 1, "atan with Ser()");

  my(f=1/2);
  for(n=1,10,
     my(got=interv_atan_Taylor([f,f],n));
     my(want=vecsort(vector(2,o, subst(Pol(atan(Ser(x,,2*(n+o)-4))),x,f))));
     check_equal(got,want));
}


\\-----------------------------------------------------------------------------
\\ interv_intersect()

{
  check_equal(interv_intersect([10,20], [15,25]), [15,20],
              "interv_intersect()");
  check_equal(interv_intersect([25,15], [20,10]), [15,20]);
  check_equal(interv_intersect([15,20], [10,20]), [15,20]);
  check_equal(interv_intersect([10,15], [10,20]), [10,15]);
  check_equal(interv_intersect([-20,10], [0,0]), [0,0]);
  
  check_equal(interv_intersect([10,20], [30,40]), 'none);
  check_equal(interv_intersect([40,30], [20,10]), 'none);
  check_equal(interv_intersect([-10,-20], [-40,-50]), 'none);
  check_equal(interv_intersect([10,20], [30,40]), 'none);
  check_equal(interv_intersect([20,10], [-5,-10]), 'none);
}

test_intersect_by_set(x,y) =
{
  x=vecsort(x);
  y=vecsort(y);
  xset = [x[1] .. x[2]];
  yset = [y[1] .. y[2]];
  tset=setintersect(xset,yset);
  if(#tset,
     [tset[1], tset[#tset]],
     'none);
}
{
  for(a=-5,5,
     for(b=-5,5,
        my(x=[a,b]);
        for(c=-5,5,
           for(d=-5,5,
              my(y=[c,d]);
              interv_intersect(x,y) == test_intersect_by_set(x,y)
              || error()))));
}

\\-----------------------------------------------------------------------------
\\ interv_sqrtnint()
{
  check_equal(interv_sqrtnint([8,8],3), [2,2]);
  check_equal(interv_sqrtnint([27,8],3), [2,3]);
  check_equal(interv_sqrtnint([124,28],3), [3,5]);
  check_equal(interv_sqrtnint([0,0],3), [0,0]);
  check_equal(interv_sqrtnint([0,1],3), [0,1]);

  check_equal(iferr(interv_sqrtnint([-1,1],3); 'no-error,
                    e,
                    errname(e)),
              "e_DOMAIN",
              "interv_sqrtnint() negative");

}

\\-----------------------------------------------------------------------------
\\ interv_sqrtint()
{
  my(r='none); 
  check_equal(sqrtint(4,&r), 2);
  check_equal(r,0);
  check_equal(interv_sqrtint([4,4]), [2,2],
              "interv_sqrtint() 4,4");
  check_equal(interv_sqrtint([9,4]), [2,3]);
  check_equal(interv_sqrtint([5,24]), [2,5]);
  check_equal(interv_sqrtint([0,0]), [0,0]);
  check_equal(interv_sqrtint([0,1]), [0,1]);

  check_equal(iferr(interv_sqrtint([-1,1]); 'no-error,
                    e,
                    errname(e)),
              "e_DOMAIN",
              "interv_sqrtint() negative");

}

\\-----------------------------------------------------------------------------
\\ interv_log_1plusx()

{
  \\ alternating series
  \\ log(1+x + O(x^6)) = x - 1/2*x^2 + 1/3*x^3 - 1/4*x^4 + 1/5*x^5
  my(x=1/19,y=1/3);
  check_equal(interv_log_1plusx([x,y],1), [-0,
                                             -0 + y],
              "interv_log_1plusx() alternating 1 terms");

  check_equal(interv_log_1plusx([x,y],3), [-0 + x - 1/2*x^2,
                                             -0 + y - 1/2*y^2 + 1/3*y^3],
              "interv_log_1plusx() alternating 3 terms");

  check_equal(interv_log_1plusx([x,y],4),
              [-0 + x - 1/2*x^2 + 1/3*x^3 - 1/4*x^4,
               -0 + y - 1/2*y^2 + 1/3*y^3]);

  my(log_term_func=(n)->-(-x)^n/n);
  check_equal(log_term_func(1), x);
  check_equal(log_term_func(2), -1/2*x^2);
  check_equal(log_term_func(3),  1/3*x^3);
  check_equal(sum(n=1,5,log_term_func(n)),
              subst(Pol(log(1+'x + O('x^6))),'x,x));

  for(num_terms=2,10,
     check_equal(interv_log_1plusx([x,x],num_terms),
                 interv_sumalt(log_term_func, num_terms, 1),
                 "interv_log_1plusx() vs interv_sumalt()"));

  \\ print(interv_log_1plusx([2,3],5) * 1.0);
  \\ print(interv_log_1plusx([2,3],10) * 1.0);
}
{
  \\ negative series
  \\ log(1-x + O(x^8))
  \\ -0 - x - 1/2*x^2 - 1/3*x^3 - 1/4*x^4 + ...

  my(x=1/17, xneg=-x);
  \\ stop at 0
  \\ -0 omit - x - 1/2*x^2 - 1/3*x^3 - 1/4*x^4 + ...
  check_equal(interv_log_1plusx([xneg,xneg],1),
              [0  - x/(1-x),
               0],
             "interv_log_1plusx() negatives 1 term");

  \\ -0 - x omit - 1/2*x^2 - 1/3*x^3 - 1/4*x^4 + ...
  check_equal(interv_log_1plusx([xneg,xneg],2),
              [-0 - x   - 1/2*x^2/(1-x),
               -0 - x ],
             "interv_log_1plusx() negatives 2 terms");

  \\ -0 - x - 1/2*x^2 omit - 1/3*x^3 - 1/4*x^4 + ...
  check_equal(interv_log_1plusx([xneg,xneg],3),
              [-0 - x - 1/2*x^2  - 1/3*x^3/(1-x),
               -0 - x - 1/2*x^2],
             "interv_log_1plusx() negatives 3 terms");
}
{
  my(t=interv_log_1plusx([1/2,1/2],6));
  \\ print("log(3/2) ",t*1.0);
  check_equal(interv_contains_constant(t,log(3/2)), 1,
             "interv_log_1plusx() contains log(3/2)");
}

\\-----------------------------------------------------------------------------
\\ interv_exp()

check_equal(exp(x + O(x^4)),  1 + x + 1/2*x^2 + 1/6*x^3 + O(x^4));
{
  check_equal(matrix(10,2,num_terms,i,num_terms--;
                     num_terms + bitnegimply(1,bitxor(num_terms,i))),
              matrix(10,2,num_terms,i,num_terms--;
                     if((num_terms%2) == (i%2),
                        num_terms+1, num_terms)));
}
check_equal(interv_exp([0,0],0), [1,1]);
{
  my(x=1/17);

  \\ 1 omit x + 1/2*x^2 + 1/6*x^3 + ...
  \\ upper bound x*(1 + x + x^2 + ...) = x* 1/(1-x)
  check_equal(x*1/(1-x), 1/16);
  check_equal(interv_exp([0,x],1), [1,
                                      1 + x/1* 1/(1 - x/1)],
             "interv_exp() 1 term");

  \\ 1 + x omit 1/2*x^2 + 1/6*x^3 + ...
  \\ upper bound x^2/2!*(1 + (x/2) + (x/2)^2 + ...)
  check_equal(interv_exp([0,x],2), [1,
                                      1+x  + x^2/2!* 1/(1 - x/2)]);

  \\ 1 + x + 1/2*x^2  omit 1/6*x^3 + ...
  \\ upper bound x^3/3! *(1 + (x/3) + (x/3)^2 + ...)
  check_equal(interv_exp([0,x],3), [1,
                                      1+x+1/2*x^2  + x^3/3!* 1/(1 - x/3)]);

  check_equal(interv_exp([0,x],3), [1,
                                      1+x+1/2*x^2  + x^3/3!* 1/(1 - x/3)]);
}
{
  my(x=1);
  check_equal(interv_exp([0,1],3), [1,
                                      1+x+1/2*x^2  + x^3/3!* 1/(1 - x/3)]);
  my(t=interv_exp([1,1],3));
  \\ print(t*1.0);
  check_equal(interv_contains_constant(t,exp(x)), 1,
             "interv_exp() contains exp(1)");
}
{
  my(x=-1);
  check_equal(interv_exp([0,-1],3), [1 + x + 1/2*x^2 + x^3/3!,
                                       1],
             "interv_exp() x negative num_terms increase");
  check_equal(interv_exp([0,-1],4), [1 + x + 1/2*x^2 + x^3/3!,
                                       1]);

  my(t=interv_exp([-1,-1],5));
  \\ print(t*1.0);
  check_equal(interv_contains_constant(t,exp(x)), 1,
             "interv_exp() contains exp(-1)");
}
{
  my(x=-2,y=-1);
  check_equal(x^3 < 0, 1);
  check_equal(y^2 > 0, 1);
  check_equal(interv_exp([x,y],3), [1 + x + 1/2*x^2 + x^3/3!,
                                      1 + y + 1/2*y^2]);

  check_equal(x^3 < 0, 1);
  check_equal(y^4 > 0, 1);
  check_equal(interv_exp([x,y],4), [1 + x + 1/2*x^2 + x^3/3!,
                                      1 + y + 1/2*y^2 + y^3/3! + y^4/4!]);
}

\\ {
\\   for(num_terms=1,10,
\\      my(t=[1/3,1/2],
\\         p=Pol(exp('x + O(x^num_terms))));
\\      \\ print(num_terms"  "p);
\\      check_equal(interv_exp(t,num_terms),
\\                  interv_subst(p,'x,t),
\\                  "interv_exp() vs subst"));
\\ }


\\------------------------------------------------------------------------------
check_equal(interv,'interv);
{
  my(test_symbols=readstr("test-symbols.txt"));
  check_equal(#test_symbols>0,1);
  for(test_i=1,#test_symbols,
     my(test_symbol=test_symbols[test_i]);
     check_equal(eval(test_symbol),
                 eval(Str("'"test_symbol)),
                 Str("leftover global variable: "test_symbol)));
}
check_equal(d,'d);
check_equal(t,'t);
check_equal(x,'x);
print("end, ",check_count," checks");
