\\ Copyright 2020, 2021, 2022, 2023, 2024 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0)
default(strictargs,1)
read("interv.gp");
read("test-helpers.gp");

option_verbose = 0;


\\-----------------------------------------------------------------------------
\\ interv_mul()

check_equal(interv_mul([1,2],[1,2]), [1,4]);
check_equal(interv_mul([-1,1],[1,1]), [-1,1]);
check_equal(interv_mul([-1,1],[-2,2]), [-2,2]);
check_equal(interv_mul([-2,3],[-5,7]), [-15,21]);
check_equal(interv_mul([2,-3],[5,-7]), [-15,21]);

test_interv_mul_by_loop(x,y) =
{
  x=vecsort(x);
  y=vecsort(y);
  my(lo=x[1]*y[1],
     hi=lo);
  for(i=x[1],x[2],
     for(j=y[1],y[2],
        lo=min(lo,i*j);
        hi=max(hi,i*j)));
  [lo,hi];
}
{
  my(ends=[-5,-2,-1,0,1,2,5]);
  for(p=1,#ends,
     for(q=1,#ends,
        for(r=1,#ends,
           for(s=1,#ends,
              my(x=[ends[p],ends[q]],
                 y=[ends[r],ends[s]]);
              check_equal(interv_mul(x,y),
                          test_interv_mul_by_loop(x,y),
                          Str("interv_mul() "x" * "y))))));
}

\\ [x,y] * t>=0  [x*t, y*t]
\\ [x,y] * t<0   [y*t, x*t]


\\-----------------------------------------------------------------------------
\\ interv_Egyptian_terms()

{
  check_equal(1/2 + 1/3,  5/6);
  check_equal(interv_Egyptian_terms([5/6,5/6]), [2,3],
             "interv_Egyptian_terms() of 5/6 exactly");
  check_equal(interv_Egyptian_terms([4+1/2,4+1/3]),
              [1,1,1,1]);
}

\\-----------------------------------------------------------------------------
\\ interv_from_Egyptian()

{
  check_equal(interv_from_Egyptian([1,1,1,1,2]),
              [4 + 1/2, 4 + 1/1]);
  check_equal(interv_from_Egyptian([2,7]),
              [1/2+1/7, 1/2+1/6],
              "interv_from_Egyptian()");
}
{
  \\ OEIS A006487  greedy Egyptian of sqrt2
  my(t_sqrt2=interv_solve_bisection([1,2], x->x^2-2, 100),
     vec=interv_Egyptian_terms(t_sqrt2,6));
  check_equal(#vec,6);
  check_equal(vec,[1, 3, 13, 253, 218201, 61323543802]);
  my(t=interv_from_Egyptian(vec));
  check_equal(interv_contains(t,t_sqrt2), 1);              
}


\\-----------------------------------------------------------------------------
\\ interv_from_Engel()

{
  \\ 1/100 was smallest denominator, so 1/99 exceeds
  check_equal(interv_from_Engel([100]),
             [1/100, 1/99]);

  check_equal(interv_from_Engel([9,13]),
             [1/9 + 1/(9*13),
              1/9 + 1/(9*12)]);
  check_equal(interv_from_Engel([9,13,17]),
             [1/9 + 1/(9*13) + 1/(9*13*17),
              1/9 + 1/(9*13) + 1/(9*13*16)]);
}
{
  \\ OEIS_data("A006784")
  my(v=[1, 1, 1, 8, 8, 17, 19, 300, 1991, 2492],
     t=interv_from_Engel(v),
     v_decr = v);
  v_decr[#v_decr]--;
  \\ print(t*1.0);
  check_equal(interv_from_Engel(v_decr)[1], t[2]);

  my(got_v=interv_Engel_terms(t));
  check_equal(#got_v, 9);
  check_equal(#got_v <= #v, 1);
  check_equal(got_v, v[1..#got_v]);

  \\ interv_from_Engel() contains actual rational
  for(rep=1,20,
     my(num=random(2^100)+1,
        den=random(2^100)+2,
        f = frac(num/den),
        v=interv_Engel_terms([f,f]));
     \\ print("len ",#v);
     check_equal(v[1] > 1, 1);
     for(len=1,#v,
        my(t=interv_from_Engel(v[1..len]));
        check_equal(interv_contains_constant(t,f), 1)));
}

\\-----------------------------------------------------------------------------
\\ interv_contfrac()

{
  check_equal(interv_contfrac(vecsort([1+1/2, 1+1/3])),
              [1]);
  check_equal(interv_contfrac(vecsort([1+1/(7 + 1/2), 1+1/(7+ 1/3)])),
              [1,7]);
  check_equal(interv_contfrac(vecsort([1+1/(7 + 1/2), 1+1/(7+ 1/3)]), 1),
              [1]);

  \\ rational has finite expansion
  my(f=quadgen(4*2));  \\ sqrt2
  check_equal(f^2 == 2, 1);
  check_equal(interv_contfrac([f,f],6),
              [1,2,2,2,2,2],
              "interv_contfrac() of quad sqrt2");

  \\ rational short endpoint
  check_equal(interv_contfrac([1/8 + 1/2^10, 1/8 + 1/2^20]),
              [0, 7, 1]);
  check_equal(contfrac(1/8 + 1/2^20),
              [0, 7, 1, 16383, 8]);
  check_equal(interv_contfrac([1/8, 1/8 + 1/2^20]),
              [0],
              "cannot distinguish 0,7,... from exact 0,8");

  \\ bug in version 2, interv_contfrac() didn't allow contfract of
  \\ one endpoint to be a prefix of the other
  check_equal(test_contfrac_to_rational([7,5,13]),    475/66, "low");
  check_equal(test_contfrac_to_rational([7,5,13,3]), 1461/203,"high");
  my(t=[475/66, 1461/203]);
  check_equal(vecsort(t), t, "is sorted");
  check_equal(interv_contfrac(t),   [7,5,13], "low shorter");
  \\
  check_equal(contfrac(54 + 12/97), [54, 8, 12], "low");
  check_equal(contfrac(54 + 1/8),   [54, 8],     "high");
  my(t=[54 + 12/97, 54 + 1/8]);
  check_equal(vecsort(t), t);
  check_equal(interv_contfrac(t),   [54, 8],  "high shorter");

  \\ rational has finite expansion
  my(f=123/457,
    t=[f,f],
    want=[0, 3, 1, 2, 1, 1, 17]);
  check_equal(contfrac(f,),       want);
  check_equal(interv_contfrac(t), want);
  check_equal(contfrac(f,,2), [0, 3]);
  check_equal(contfrac(f,,3), [0, 4]);
  check_equal(contfrac(f,,4), [0, 3, 1, 2]);
  check_equal(vector(20,maxlen,maxlen--; 
                     #interv_contfrac(t,maxlen)),
              vector(20,maxlen,maxlen--; min(7,maxlen)),
              "interv_contfrac() maxlen");
}

\\-----------------------------------------------------------------------------
\\ interv_Engel_terms()

{
  check_equal(interv_Engel_terms([1/2,1/2]),
              [2],
             "interv_Engel_terms() of 1/2 exactly");

  check_equal(interv_Engel_terms([1/2,
                                  1/2 +1/10^100],4),
              [2],
              "interv_Engel_terms() of 1/2 exactly");
  
  my(bits=100,
     p=2*4^bits,
     lo=sqrtint(p),
     hi=lo+1,
     t=[lo/2^bits, hi/2^bits]);
  check_equal(interv_Engel_terms(t,8),
              [1, 3, 5, 5, 16, 18, 78, 102]); \\ A028254
}


\\-----------------------------------------------------------------------------
\\ interv_from_contfrac()

{
  check_equal(contfracpnqn(['x,'y,'z]),  
              [(z*y + 1)*x + z,         y*x + 1;
               z*y + 1,                 y]);
  check_equal(contfracpnqn(['x,'y,'z,1]), 
              [((z+1)*y + 1)*x + z+1, (z*y + 1)*x + z;
               (z+1)*y + 1,             z*y + 1]);
  check_equal((z*y + 1)*x + z + y*x + 1, 
              ((z+1)*y + 1)*x + z+1);
  check_equal(z*y + 1 + y,  
              (z+1)*y + 1);

  check_equal(interv_from_contfrac([0]), [0,1]);
  check_equal(interv_from_contfrac([6]), [6,7]);
  my(check(vec) =
     my(want=vecsort([test_contfrac_to_rational(vec),
                      test_contfrac_to_rational(concat(vec,1))]));
     check_equal(interv_from_contfrac(vec), want));

  check([32,432,16,20]);
  check([32,432,1]);
  check([0]);

  for(rep=1,50,
     my(len=random(8)+1,
        vec=vector(len,i, random(20)+1));
     check(vec));

  check_equal(test_contfrac_to_rational([1,1,1]),   45/30);
  check_equal(test_contfrac_to_rational([1,1,1,7]), 46/30);

  check_equal(test_contfrac_to_rational([1,1]),   16/8);
  check_equal(test_contfrac_to_rational([1,1,7]), 15/8);

  \\ next term makes bigger or smaller according to parity of length
  check_equal(vector(20,n,
                     my(v=vector(n,i,1),
                        r=test_contfrac_to_rational(v),
                        s=test_contfrac_to_rational(concat(v,7)));
                     r > s),
              vector(20,n, n%2==0));
}


\\-----------------------------------------------------------------------------
\\ interv_widen_integer()

check_equal(interv_widen_integer([1/3,1/3]), [0,1]);
check_equal(interv_widen_integer([1/3,-2/3]), [-1,1]);


\\-----------------------------------------------------------------------------
\\ interv_div()

{
  check_equal(interv_div([13,17],[4,4]), [13/4,17/4]);
  check_equal(interv_div([13,17],[2,4]), [13/4,17/2]);
  check_equal(interv_div([13,17],[-2,-4]), [-17/2,-13/4]);
  check_equal(iferr(interv_div([13,17],[-2,4]); 'no-error,
                    e,
                    errname(e)),
              "e_INV",
              "interv_div() divide by 0");
}

\\-----------------------------------------------------------------------------
\\ interv_add()

{
  check_equal(interv_add([1,2],[4,5]), [5,7],
              "1..2 + 4..5");
  check_equal(interv_add([1,-2],[4,5]), [2,6]);
}


\\-----------------------------------------------------------------------------
\\ interv_add_constant()

check_equal(interv_add_constant([1,2],9), [10,11]);
check_equal(interv_add_constant([1,-2],9), [7,10]);


\\-----------------------------------------------------------------------------
\\ interv_sub_constant()

check_equal(interv_sub_constant([1,2],9), [-8,-7]);
check_equal(interv_sub_constant([1,-2],9), [-11,-8]);


\\-----------------------------------------------------------------------------
\\ interv_union()

{
  check_equal(interv_union([1,2],[4,5]), [1,5]);
  check_equal(interv_union([-2,1],[4,5]), [-2,5]);
  check_equal(interv_union([1,2],[5,-4]), [-4,5]);
}

\\-----------------------------------------------------------------------------
\\ interv_union_0()

check_equal(interv_union_0([1,5]), [0,5]);
check_equal(interv_union_0([-1,-5]), [-5,0]);
check_equal(interv_union_0([4,-5]), [-5,4]);


\\-----------------------------------------------------------------------------
\\ interv_contains_0()

check_equal(interv_contains_0([-5,4]),1);
check_equal(interv_contains_0([4,-5]),1);
check_equal(interv_contains_0([-1,-5]),0);
check_equal(interv_contains_0([5,1]),0);
check_equal(interv_contains_0([0,1]),1);
check_equal(interv_contains_0([0,-1]),1);


\\-----------------------------------------------------------------------------
\\ interv_contains_constant()

check_equal(interv_contains_constant([-5,4],-10),0);
check_equal(interv_contains_constant([-5,1],-2),1);
check_equal(interv_contains_constant([-5,1],10),0);

check_equal(interv_contains_constant([4,-5],-10),0);
check_equal(interv_contains_constant([1,-5],-2),1);
check_equal(interv_contains_constant([1,-5],10),0);


\\-----------------------------------------------------------------------------
\\ interv_pow_constant()

\\ plot(x=-3,-2, x^2)
check_equal(interv_pow_constant([-3,-2],2), [4,9]);

check_equal(interv_pow_constant([2,3],5), [2^5, 3^5]);
check_equal(interv_pow_constant([2,3],-1), [1/3,1/2]);
check_equal(interv_pow_constant([-3,-2],5), [-3^5,-2^5]);
check_equal(interv_pow_constant([-1,1],1), [-1,1]);
check_equal(interv_pow_constant([-1,1],2), [0,1]);
check_equal(interv_pow_constant([-1/2,-1/2],2), [1/4,1/4]);
check_equal(interv_pow_constant([-1/2,1/2],2), [0,1/4]);
check_equal(interv_pow_constant([-1/2,1/2],3), [-1/8,1/8]);
check_equal(interv_pow_constant([1/3,1/2],2), [1/9,1/4]);

check_equal(interv_pow_constant([2,3],2),     [4,9]);
check_equal(interv_mul([2,3],[2,3]), [4,9]);
check_equal(interv_pow_constant([2,3],3),     [8,27]);
check_equal(interv_mul([4,9],[2,3]), [8,27]);

\\ test_interv_pow_constant_by_mul(t,n) =
\\ {
\\   n>=1 || error();
\\   my(ret=t);
\\   for(i=2,n, ret=interv_mul(ret,t));
\\   ret;
\\ }
\\ {
\\   my(ends=[-5,-2,-1,0,1,2,5]);
\\   for(p=1,#ends,
\\      for(q=1,#ends,
\\         for(n=2,6,
\\            my(x=[ends[p],ends[q]]);
\\            check_equal(interv_pow_constant(x,n),
\\                        test_interv_pow_constant_by_mul(x,n),
\\                        Str("interv_pow_constant() "x" ^ "n)))));
\\ }


\\-----------------------------------------------------------------------------
\\ interv_frac_digits()

check_equal(interv_frac_digits([1/3,1/2]), [0]);
check_equal(interv_frac_digits([(97+1/3)/100,(97+1/2)/100]), [0,9,7]);


\\-----------------------------------------------------------------------------
\\ interv_frac_fromdigits()

check_equal(interv_frac_fromdigits([0]), [0,1]);


\\-----------------------------------------------------------------------------
\\ interv_solve_bisection()

check_equal(interv_solve_bisection([-2,1], t->2*t, 1), [-1/2,1]);
check_equal(interv_solve_bisection([-2,1], t->2*t, 2), [-1/2,1/4]);
check_equal(interv_solve_bisection([0,3], t->t^2-1, 1), [0,3/2]);


\\-----------------------------------------------------------------------------
\\ interv_sub()

check_equal(interv_sub([100,110],[5,7]), [ 93, 105]);


\\-----------------------------------------------------------------------------
\\ interv_recip()

check_equal(interv_recip([2,3]), [1/3,1/2]);
check_equal(interv_recip([-3,-2]), [-1/2,-1/3]);

\\ iferr ...
\\ check_equal(interv_recip([-1,1]), [1,oo]);
\\ check_equal(interv_recip([-3,1]), [1/3,oo]);
\\ check_equal(interv_recip([-1,3]), [1/3,oo]);

check_equal(interv_recip([-7,-5]), [-1/5, -1/7]);
check_equal(interv_recip([5,7]), [1/7, 1/5]);
check_equal(interv_recip([-7,-1]), [-1,-1/7]);
check_equal(interv_recip([1,7]), [1/7, 1]);
check_equal(interv_recip([-7,-1/3]), [-3,-1/7]);
{
  for(i=-5,5,
     for(j=-5,5,
        my(t=[i,j]);
        if(interv_contains_0(t),next);
        my(got=interv_recip(t));
        check_equal(got[1] <= got[2], 1)));
}


\\-----------------------------------------------------------------------------
\\ interv_neg()

check_equal(interv_neg([1,2]), [-2,-1]);
check_equal(interv_neg([2,1]), [-2,-1]);
check_equal(interv_neg([-1,-2]), [1,2]);
check_equal(interv_neg([-2,-1]), [1,2]);
check_equal(interv_neg([-1,1]), [-1,1]);
check_equal(interv_neg([1,-1]), [-1,1]);
{
  for(i=-5,5,
     for(j=-5,5,
        my(t=[i,j],
           got=interv_neg(t));
        check_equal(got[1] <= got[2], 1)));
}


\\-----------------------------------------------------------------------------
\\ interv_widen_bits()

check_equal(interv_widen_bits([1/3,1/2],0), [0,1]);
check_equal(interv_widen_bits([1/3,1/2],4), [5/16,1/2]);
check_equal(5/16 < 1/3, 1);
check_equal(interv_contains(    [5/16,1/2],   [1/3,1/2]), 1);


\\------------------------------------------------------------------------------
check_equal(interv,'interv);
{
  my(test_symbols=readstr("test-symbols.txt"));
  check_equal(#test_symbols>0,1);
  for(test_i=1,#test_symbols,
     my(test_symbol=test_symbols[test_i]);
     check_equal(eval(test_symbol),
                 eval(Str("'"test_symbol)),
                 Str("leftover global variable: "test_symbol)));
}

\\-----------------------------------------------------------------------------
check_equal(d,'d);
check_equal(n,'n);
check_equal(t,'t);
check_equal(x,'x);
print("end");
