\\ Copyright 2022 Kevin Ryde
\\ 
\\ This file is part of OEIS-data.
\\
\\ OEIS-data is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the
\\ Free Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ OEIS-data is distributed in the hope that it will be useful, but WITHOUT
\\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
\\ FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
\\ more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with OEIS-data.  If not, see <http://www.gnu.org/licenses/>.


\\ Usage: read("OEIS-data.gp");
\\        print(OEIS_data("A123456"));
\\
\\ This is some code for reading sequence data and related information from
\\ downloaded files.
\\
\\     vec = OEIS_data(anum,flags=0)
\\     offset = OEIS_offset(anum)
\\     [offset,vec] = OEIS_offset_and_data(anum,flags=0)
\\     func = OEIS_data_func(anum,flags=0)
\\     ser = OEIS_data_gf(anum,flags=0)
\\     str = OEIS_name(anum)
\\
\\
\\ A-Numbers
\\ =========
\\
\\ An "anum" parameter can be a string or a symbol.  Quote the symbol in the
\\ usual way in case the symbol is defined as a function etc.
\\
\\     OEIS_data("A123456")   \\ string arg
\\     OEIS_data('A123456)    \\ symbol arg
\\
\\
\\ Data Files
\\ ----------
\\
\\ Information is taken from one or more of the following files, in
\\ directory ~/OEIS, where ~ means the home directory (as determined by Perl
\\ module File::HomeDir currently).
\\
\\     ~/OEIS/A123456.internal.txt
\\     ~/OEIS/b123456.txt
\\     ~/OEIS/stripped
\\     ~/OEIS/names
\\
\\ These are as downloaded for instance from, respectively,
\\
\\     http://oeis.org/search?q=id:A123456&fmt=text
\\     http://oeis.org/A123456/b123456.txt
\\     http://oeis.org/stripped.gz         # and then gunzipped
\\     http://oeis.org/names.gz            # and then gunzipped
\\
\\ The "text" internal format is used since it's a machine-readable form.
\\
\\ The stripped file is large but is handy for having all DATA sections.
\\ It doesn't include OFFSETs so itindividual A123456.internal.txt files are
\\ needed for that.  (Or an 'offset in the flags parameters here.)
\\
\\ The names file is large but is handy for all NAME sections.  Nothing here
\\ depends on having the names, but they can be useful in printouts for
\\ human readers.
\\
\\ Files tried for information combinations include:
\\
\\     OEIS_data()
\\         internal.txt
\\         OR stripped
\\         OR b-file if rquested by flags
\\
\\     OEIS_offset() 
\\         internal.txt
\\         OR b-file
\\
\\     OEIS_offset_and-data()
\\     OEIS_offset_func()
\\     OEIS_offset_gf()
\\         internal.txt
\\         OR stripped and b-file
\\         OR b-file if rquested by flags
\\
\\     OEIS_name()
\\         internal.txt
\\         OR names
\\
\\ b-files are sometimes overly large.  A truncated file is fine as long as
\\ the format is right, and of course ends with a complete last line -- not
\\ chopped in the middle of a sequence value.
\\
\\ The files have partly overlapping information.  Be sure they're
\\ consistent.
\\
\\
\\ Data Values
\\ -----------
\\ 
\\ The simplest function is
\\ 
\\     vec = OEIS_data("A123456")
\\
\\ which returns a vector of the DATA section from sequence A123456.  The
\\ values from b-file b123456.txt can be read with the 'bfile parameter
\\
\\     vec = OEIS_data("A123456",'bfile)
\\     
\\ These are often enough for inspecting or manipulating, but variations
\\ here include a function form.  It can be assigned to any name,
\\
\\     func = OEIS_data_func("A123456")
\\     \\ to be called func(n)
\\     
\\ This requires the OFFSET, as per OEIS_offset_and_data().  It can be
\\ supplied in the flags parameter if you have the stripped file but keep
\\ only a selection of internal.txt files.
\\ 
\\     func = OEIS_data_func("A123456",['offset,1])
\\     
\\ In "cons" sequences, the offset is how many digits of the constant are
\\ before the decimal point.  Here cons sequencess are treated the same any
\\ other with func(o), func(o+1), etc where o is the offset.
\\ 
\\ It's been pointed out before that this is not particularly helpful for
\\ operating on digits, but don't want to go against the usual offsets way,
\\ and some sequences have a dual nature of say periodic pattern and cons of
\\ a rational.
\\ 
\\ Like the plain OEIS_data(), values can be taken from the b-file,
\\
\\     func = OEIS_data_func("A123456",'bfile)
\\     \\ using b123456.txt
\\ 
\\ Data values can be had in the form of a series (t_SER) which is useful
\\ for generating function manipulations.
\\     
\\     gf = OEIS_data_gf("A123456")
\\     gf = OEIS_data_gf("A123456",'bfile)
\\     
\\     series  a(0)*x^0 + x(1)*x^1 + ... + O(x^len)
\\     
\\ The series precision O(x^len) is after the end of the data (or b-file).
\\ Series calculations propagate and combine such ends so that acalculatin
\\ in generating functions ends with as much as is certain from the data.
\\     
\\     
\\ Locale Charset
\\ --------------
\\
\\ OEIS_name() returns a string of bytes in the locale charset.  This is
\\ normally what's wanted for a printout.  The locale charset is determined
\\ by the Perl Encode::Locale module if you have that and new enough Perl,
\\ or if not then utf8 is assumed (that being the coding of the OEIS files
\\ themselves too).
\\
\\
\\ Internals
\\ ---------
\\ 
\\ Functions and variables named OEIS_INTERNAL_foo are for internal-use (as
\\ the name suggests) and will change wildly.
\\
\\
\\ Limitations
\\ -----------
\\ 
\\ The data reading depends on having various downloaded files, possibly a
\\ lot of files if working on several sequence.  The download internal.txt
\\ are in the style of my Emacs oeis.el for getting files one by one.  Some
\\ automated downlaod within the OEIS-data-reader program here could be
\\ convenient.  Maybe with a limit on b-file length.
\\ 
\\ One main intended use is to check functions or calculations against OEIS
\\ data values, as a mutual check of code against reference data.  Some
\\ checks are in OEIS-data-wip.gp but are not quite ready to call stable.
\\ 
\\
\\ Changes
\\ -------
\\
\\ Version 1 - the first version.

OEIS_data_version = 1;

\\ Perl script which does actual data reading.
OEIS_INTERNAL_program = "OEIS-data-reader";

{
\\ if not already set by the user
if(OEIS_check_verbose=='OEIS_check_verbose,
OEIS_check_verbose = 0
);
}

\\-----------------------------------------------------------------------------

\\ Require A123456 or 7-digit A1234567 looking ahead to a (slightly) distant
\\ future passing a million sequences.
\\
my(allow=Set(Vecsmall("A0123456789"))); \
OEIS_INTERNAL_validate_anum(anum) =
{
  anum=Str(anum);
  ((#anum == 6 || #anum == 7)
  && !#setminus(Set(Vecsmall(anum)),allow))
    || error("Bad A-number: ",anum);
  anum;
}

\\ Require alphanumerics - and _ so as not to run a shell command with wild
\\ and possibly dangerout chars.
\\
my(allow=Set(Vecsmall("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"))); \
OEIS_INTERNAL_validate_flag(flag) =
{
  flag=Str(flag);
  !#setminus(Set(Vecsmall(flag)),allow)
    || error("Bad flag: ",flag);
  flag;
}

\\ program run like
\\     OEIS-data-reader --check-version 1 \
\\                      --anum A123456 \
\\                      --action offset_and_data \
\\                      --flag offset
\\                      --flag 0
\\
OEIS_INTERNAL_run(anum,action,flags) =
{
  my(str=Str(OEIS_INTERNAL_program,
             " --check-version ",OEIS_data_version
             " --anum ",OEIS_INTERNAL_validate_anum(anum),
             " --action ",action));
  if (type(flags) != "t_VEC", flags=[flags]);
  for(i=1,#flags,
     str = Str(str,
               " --flag ",OEIS_INTERNAL_validate_flag(flags[i])));
  \\ print(str);
  extern(str);
}

OEIS_name(anum) = OEIS_INTERNAL_run(anum,"name",0);
{
addhelp(OEIS_name,
"str = OEIS_name(anum)
Return a string which is the name of sequence anum, or string \"unknown\" if
the name is not known.

The name might contain non-ASCII characters and these are returned as bytes
in the current locale charset if that can be determined (or UTF-8 otherwise).");
}

OEIS_data(anum,flags=0) = OEIS_INTERNAL_run(anum,"data",flags);
{
addhelp(OEIS_data,
"vec = OEIS_data(anum,flags=0)
Return a vector of the DATA section from anum.

Optional flags parameter can be symbol 'bfile to get the data from
b123456.txt instead, which might be much larger.");
}

OEIS_offset(anum) = OEIS_INTERNAL_run(anum,"offset",0);
{
addhelp(OEIS_offset,
"offset = OEIS_offset(anum)
Return the \"offset\" of OEIS sequence anum.
This is the index \"n\" for the first term of the sequence data, or for a
\"cons\" sequence the number of decimal digits in the data before the
decimal point (eg. 2 for 13.1415926...).");
}

OEIS_offset_and_data(anum,flags=0) = \
  OEIS_INTERNAL_run(anum,"offset_and_data",flags);
{
addhelp(OEIS_offset_and_data,
"[offset,vec] = OEIS_offset_and_data(anum,flags=0)
Return a vector of two elements: the offset (integer) and data vector of
anum, as per OEIS_offset() and OEIS_data().");
}

OEIS_data_func(anum,flags=0) =
{
  my(offset,data);
    [offset,data] = OEIS_offset_and_data(anum,flags);
  (n)-> n>=offset || error(anum," is for n>=",offset);
        n < offset+#data || error(anum," data extends only to n<=",offset+#data-1);
        data[n-offset+1];
}
{
addhelp(OEIS_data_func,
"func = OEIS_data_func(anum,flags=0)
Return a function which can be called func(n) to get terms of anum.
Optional flag parameter can be symbol 'bfile to get the data from a b-file.

flags can be a vector ['offset,123] to specify the sequence offset (if you
only have the stripped file).");
}

OEIS_data_gf(anum,flags=0) =
{
  my(offset,data); 
    [offset,data] = OEIS_offset_and_data(anum,flags);
  Ser(data,,#data) * 'x^offset;
}
{
addhelp(OEIS_data_gf,
"ser = OEIS_data_gf(anum,flags=0)
Return a t_SER series of a(n)*x^n, with the series \"precision\" to
O(x^after) which is after the sequence data.

Optional parameter flags can be symbol 'bfile to get b-file data, and can be
['offset,123] to specify the sequence OFFSET if only the stripped file..");
}

\\-----------------------------------------------------------------------------
