#!/usr/bin/perl -w

# Copyright 2017, 2018, 2019, 2021, 2022 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Miscellaneous consistency checks for interv.gp and interv-wip.gp.

use strict;
use FindBin;
use File::Slurp;
use File::Spec;
use Tie::IxHash;
use Regexp::Common 'delimited','comment';

# uncomment this to run the ### lines
# use Smart::Comments;

# match a GP string, including any backslash escaping in it
my $gp_string_re = qr{$RE{delimited}{-delim=>'"'}{-esc=>'\\'}}o;
my $gp_comment_re = qr/\\\\.*|$RE{comment}{C}/;

sub str_strip_strings_and_comments {
  my ($str) = @_;
  $str =~ s{(?<eval>   eval\(        )?
            (?<string> $gp_string_re )
            |
            $gp_comment_re
           }
           {my $string = $+{'string'};
            my $s;
            if ($+{'eval'}) {
              $s = $string;    # eval("string") keep string content as code
              $s =~ s/^"//;
              $s =~ s/"$//;
              ### eval: $s
            } else {
              $s = $&;
              $s =~ tr/\n//cd;      # keep newlines only
              if (defined $string) { $s = "\"$s\""; }
            }
            $s
          }egx;
  return $str;
}
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

my $bad = 0;
sub bad {
  print @_;
  $bad++;
}

my $srcdir = File::Spec->catdir($FindBin::Bin, File::Spec->updir);
my $examples_dir = File::Spec->catdir($srcdir, 'examples');

# All functions defined by the code.
#
tie my %funcs, 'Tie::IxHash';
my %examples;

sub check_filename {
  my ($filename, %options) = @_;
  ### $filename
  my $str = File::Slurp::read_file($filename);
  my $str_sans_comments = str_strip_strings_and_comments($str);

  # all function definitions
  #
  while ($str_sans_comments =~
         /^((?<func>interv_\w+)(?<args>\(.*?\))           # funcname(args) = code
          |(?<func>interv_\w+)\ *\=\ *(?<args>\(.*?\))->  # funcname = (x)->code
          |(?<func>interv_\w+)\ \=                        # varname = (x)->code
          |kill\((?<func>interv_\w+)
          )/mgx) {
    my $func = $+{'func'};
    my $args = $+{'args'};
    my $linenum = pos_linenum($str_sans_comments);
    ### $func
    ### $args
    if (exists $funcs{$func}) {
      bad("$filename:$linenum: duplicate definition: $func\n");
    } else {
      $funcs{$func} = { args => $args,
                        filename => $filename,
                        linenum => $linenum,
                      };
    }
  }

  # All "interv_*" appearing, other than in comments, must be a defined
  # function.
  #
  my $count_words = 0;
  while ($str_sans_comments =~ /\binterv_[A-Za-z0-9_]+/g) {
    my $func = $&;
    $count_words++;
    unless (exists $funcs{$func}) {
      my $linenum = pos_linenum($str_sans_comments);
      bad("$filename:$linenum: $func undefined\n");
    }
  }

  # Prefer sqr() over ^2.
  while ($str_sans_comments =~ /\^2/g) {
    my $linenum = pos_linenum($str_sans_comments);
    bad("$filename:$linenum: prefer sqr() over ^2\n");
  }

  # All interv_* listed in the initial summary comments must be defined functions.
  #
  # start of file through to "Notes" heading
  $str =~ /This is some code.*?Intervals|Work in progress.*?End Summary/s
    or die "Oops, no match summary in $filename";
  my $summary_str = $&;
  my $summary_pos = $-[0];
  ### $summary_pos

  while ($summary_str =~ /^\\\\\ {5}                     # indented
                          (?<ret>  [[:graph:]]*\ =\ )?   # optional return value
                          (?<func> interv_\w+)
                          (?<args> \(.*?\))?
                         /mgx) {
    my $ret  = $+{'ret'};      # return value, including "="
    my $func = $+{'func'};     # function name
    my $args = $+{'args'} // "[no args]";
    my $linenum = pos_linenum($str, pos($summary_str)+$summary_pos);
    ### $ret
    ### $func
    ### $args

    my $h = $funcs{$func};
    if (! $h) {
      bad("$filename:$linenum: summary func undefined: $func\n");
      next;
    }
    if (defined $h->{'summary_usage'}) {
      bad("$filename:$linenum: duplicate in summary: $func\n");
    }
    $h->{'summary_usage'} = (defined $ret ? "$ret$func$args" : "$func$args");

    my $definition_args = $h->{'args'} // '[no args]';
    unless ($args eq $definition_args) {
      bad("$filename:$linenum: summary $func args: $args\n");
      print "$h->{filename}:$h->{linenum}: ... but definition: $definition_args\n";
    }
  }

  # All defined functions must be in the summary.
  #
  foreach my $func (keys %funcs) {
    next if $func =~ /INTERNAL/;     # internal funcs not in summary
    my $h = $funcs{$func};
    unless (defined $h->{'summary_usage'}) {
      bad("$h->{filename}:$h->{linenum}: not in summary: $func\n");
    }
  }

  # Each function which exists must have an addhelp.
  # In each addhelp string the functions referred to must exist.
  #
  while ($str =~ /^addhelp\(
                  (?<func>\w+)
                  (.*?\n(?<string>$gp_string_re))?
                 /mgx) {
    my $func   = $+{'func'};
    my $string = $+{'string'};
    my $pos    = $-[0];

    if (! defined $string) {
      my $linenum = pos_linenum($str, $pos);
      bad("$filename:$linenum: addhelp $func no string part\n");
      next;
    }
    my $h = $funcs{$func};
    if (! $h) {
      my $linenum = pos_linenum($summary_str, $pos);
      bad("$filename:$linenum: addhelp func undefined: $func\n");
      next;
    }

    if ($h->{'seen_addhelp'}++) {
      my $linenum = pos_linenum($summary_str, $pos);
      bad("$filename:$linenum: duplicate addhelp: $func\n");
    }

    my $usage = $string;     # first line of string part
    $usage =~ s/^"//;
    $usage =~ s/\n.*//s;
    if (defined $h->{'summary_usage'} && $usage ne $h->{'summary_usage'}) {
      my $linenum = pos_linenum($str, $pos);
      bad(  "$filename:$linenum: addhelp usage different: $usage\n");
      print "$filename:$linenum: ... from summary:        $h->{summary_usage}\n";
    }

    while ($string =~ /\binterv_\w+/g) {
      my $func = $&;
      next if $func =~ /_$/;
      unless (exists $funcs{$func}) {
        my $linenum = pos_linenum($str, $pos+pos($string)) + 1;
        bad("$filename:$linenum: addhelp string contains undefined function: $func\n");
      }
    }
  }

  # example programs mentioned anywhere
  while ($str =~ m{\bexamples/(?<filename>.*?\.gp\b)}g) {
    my $example = $+{'filename'} // next;
    my $linenum = pos_linenum($str);
    $examples{$example} //= "$filename:$linenum";
    if (! -e File::Spec->rel2abs($example, $examples_dir)) {
      bad("$filename:$linenum: no such example program: $example\n");
    }    
  }

  # no leftover diagnostic print()
  while ($str_sans_comments =~ m{\bprint1?}g) {
    my $print = $&;
    my $linenum = pos_linenum($str_sans_comments);
    bad("$filename:$linenum: leftover diagnostic $print\n");
  }

  print "$filename ",scalar(keys %funcs)," defines, $count_words words\n";
}

sub check_tests {
  my ($test_filename) = @_;
  my $str = File::Slurp::read_file($test_filename);
  my $str_sans_comments = str_strip_strings_and_comments($str);

  foreach my $func (keys %funcs) {
    next if $func =~ /INTERNAL/;
    next if $funcs{$func}->{'tested'};
    if ($str_sans_comments =~ /\b\Q$func\E\b/) {
      $funcs{$func}->{'tested'} = 1;
    } else {
      bad("$test_filename:1: no test of $func\n");
    }
  }
}

sub check_example_program {
  my ($filename) = @_;
  my $str = File::Slurp::read_file($filename);
  my ($volume,$directories,$basename) = File::Spec->splitpath($filename);

  unless ($str =~ /Usage:.*\b\Q$basename\E\b/) {
    bad("$filename:1: no Usage comment\n");
  }
}

check_filename(File::Spec->catfile($srcdir,'interv.gp'));

{
  my $count = 0;
  foreach my $filename (glob(File::Spec->catfile($examples_dir,'*.gp'))) {
    $count++;
    $filename = File::Spec->abs2rel($filename);
    my ($volume,$directories,$basename) = File::Spec->splitpath($filename);
    unless ($examples{$basename}
            || $basename eq 'bisection.gp'
           ) {
      bad("$filename:1: not mentioned in interv.gp\n");
    }
    check_example_program($filename);
  }
  print "$count example programs\n";
}

check_tests(File::Spec->catfile($srcdir,'test.gp'));

check_filename(File::Spec->catfile($srcdir,'interv-wip.gp'));
# tests of interv-wip.gp are optional
# check_tests(File::Spec->catfile($srcdir,'test-wip.gp'));

exit ($bad ? 1 : 0);
