#!/usr/bin/perl -w

# Copyright 2017, 2018, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl my-gp-symbols.pl [--options] filename.gp...
#
# Print symbols used in the given Pari/GP script file or files.

use 5.010;
use warnings;
use strict;
use File::Slurp;
use FindBin;
use Getopt::Long;
use Regexp::Common 'comment','delimited';
use IPC::Run;
use List::Util 'max','min';
use Tie::IxHash;

# uncomment this to run the ### lines
# use Smart::Comments;

our $VERSION = 0;

my $progname = $FindBin::Script;
my $verbose = 0;
my @exclude_prefixes;
{
  my $help = sub {
    print "$progname [--options] filename...\n";
    my @opts =
      (['-h, --help',    'Print this help'],
       ['-v, --version', 'Print program version'],
       ['--verbose',     'Print extra messages'],
      );
    my $width = 2 + max (map { length ($_->[0]) } @opts);
    foreach (@opts) {
      printf "%-*s%s\n", $width, $_->[0], $_->[1];
    }
    print "\n";
    exit 0;
  };

  GetOptions ('exclude-prefix:s' => sub {
                my ($opt,$value) = @_;
                push @exclude_prefixes, $value;
              },
              'help|?' => $help,
              'verbose:i'    => \$verbose,
              version => sub {
                print "$progname version $VERSION\n";
                exit 0;
              },
             )
    or exit 1;

  @ARGV or $help->();
}

if ($verbose) {
  print "exclude prefixes ",join(' ',@exclude_prefixes),"\n";
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

# match a GP word
my $gp_word_re = qr/[A-Za-z_][A-Za-z_0-9]*/;

# match a GP comment,
#    /* ... */
#    \\ to newline
# my $gp_comment_re = qr{/\*.*?\*/|\\\\[^\r\n]*};
my $gp_comment_re = qr{$RE{comment}{C}|\\\\[^\r\n]*}o;

# '/* GP-Test */' =~ $gp_comment_re or die;

# match a GP string, including any backslash escaping in it
# my $gp_string_re = qr{"(\\.|[^\\"])*"?};
my $gp_string_re = qr{$RE{delimited}{-delim=>'"'}{-esc=>'\\'}}o;


'"xx"'=~ $gp_string_re or die;
"\"x\\\"x\""=~ $gp_string_re or die;


use constant::defer builtin_symbols_href => sub {
  my $str = '';
  # gphelp in 2.11.0 takes space ' ' to mean an operator.
  # 'k' matches \kbdsidx macro in raw text of helps.
  foreach my $arg (' ','k') {
    IPC::Run::run(['gphelp','-k','k'],
                  '<', File::Spec->devnull,
                  '>>', \$str,
                  '2>', File::Spec->devnull);
  }
  $str =~ s/See also:.*//;
  $str =~ s/\_/_/;
  my @words = split /\s+/, $str;
  @words = grep $gp_word_re, @words;  # not $L$-functions and operators
  my %words = (my => 1,  # not reported by gphelp
              );
  @words{@words} = (); # hash slice
  return \%words;
};

my %symbols;
tie %symbols, 'Tie::IxHash';

sub file_symbols {
  my ($filename) = @_;
  my $str = File::Slurp::slurp($filename);
  if ($verbose) {
    print "$filename ",count_lines($str)," lines\n";
  }

 STR: while ($str =~ m{(?<Comment> $gp_comment_re)
                     |(?<String> $gp_string_re)
                     |(?<Word> $gp_word_re)}xg) {
    # if (defined $+{'String'}) { print STDERR $+{'String'}; }
    my $word = $+{'Word'} // next;
    ### $word

    foreach my $exclude_prefix (@exclude_prefixes) {
      next STR if $word =~ /^\Q$exclude_prefix/;
    }
    next if exists builtin_symbols_href()->{$word};
    $symbols{$word}++;
  }
}

foreach my $filename (@ARGV) {
  file_symbols($filename);
}

foreach my $symbol (keys %symbols) {
  print $symbol,"\n";
}

exit 0;
